from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase

from explorer.apps import _validate_connections


class TestApps(TestCase):

    @patch('explorer.apps._get_default')
    def test_validates_default_connections(self, mocked_connection):
        mocked_connection.return_value = 'garbage'
        self.assertRaises(ImproperlyConfigured, _validate_connections)

    @patch('explorer.apps._get_explorer_connections')
    def test_validates_all_connections(self, mocked_connections):
        mocked_connections.return_value = {'garbage1': 'in', 'garbage2': 'out'}
        self.assertRaises(ImproperlyConfigured, _validate_connections)



import io
import random
import string
from datetime import date, datetime, timedelta

from django.core.cache import cache
from django.core.mail import send_mail

from explorer import app_settings
from explorer.exporters import get_exporter_class
from explorer.models import Query, QueryLog


if app_settings.ENABLE_TASKS:
    from celery import shared_task
    from celery.utils.log import get_task_logger

    from explorer.utils import s3_upload
    logger = get_task_logger(__name__)
else:
    import logging

    from explorer.utils import noop_decorator as shared_task
    logger = logging.getLogger(__name__)


@shared_task
def execute_query(query_id, email_address):
    q = Query.objects.get(pk=query_id)
    send_mail('[SQL Explorer] Your query is running...',
              f'{q.title} is running and should be in your inbox soon!',
              app_settings.FROM_EMAIL,
              [email_address])

    exporter = get_exporter_class('csv')(q)
    random_part = ''.join(
        random.choice(
            string.ascii_uppercase + string.digits
        ) for _ in range(20)
    )
    try:
        url = s3_upload(f'{random_part}.csv', convert_csv_to_bytesio(exporter))
        subj = f'[SQL Explorer] Report "{q.title}" is ready'
        msg = f'Download results:\n\r{url}'
    except Exception as e:
        subj = f'[SQL Explorer] Error running report {q.title}'
        msg = f'Error: {e}\nPlease contact an administrator'
        logger.exception(f'{subj}: {e}')
    send_mail(subj, msg, app_settings.FROM_EMAIL, [email_address])


# I am sure there is a much more efficient way to do this but boto3 expects a binary file basically
def convert_csv_to_bytesio(csv_exporter):
    csv_file_io = csv_exporter.get_file_output()
    csv_file_io.seek(0)
    csv_data: str = csv_file_io.read()
    bio = io.BytesIO(bytes(csv_data, 'utf-8'))
    return bio


@shared_task
def snapshot_query(query_id):
    try:
        logger.info(f"Starting snapshot for query {query_id}...")
        q = Query.objects.get(pk=query_id)
        exporter = get_exporter_class('csv')(q)
        k = 'query-{}/snap-{}.csv'.format(
            q.id,
            date.today().strftime('%Y%m%d-%H:%M:%S')
        )
        logger.info(f"Uploading snapshot for query {query_id} as {k}...")
        url = s3_upload(k, convert_csv_to_bytesio(exporter))
        logger.info(
            f"Done uploading snapshot for query {query_id}. URL: {url}"
        )
    except Exception as e:
        logger.warning(
            f"Failed to snapshot query {query_id} ({e}). Retrying..."
        )
        snapshot_query.retry()


@shared_task
def snapshot_queries():
    logger.info("Starting query snapshots...")
    qs = Query.objects.filter(snapshot=True).values_list('id', flat=True)
    logger.info(
        f"Found {len(qs)} queries to snapshot. Creating snapshot tasks..."
    )
    for qid in qs:
        snapshot_query.delay(qid)
    logger.info("Done creating tasks.")


@shared_task
def truncate_querylogs(days):
    qs = QueryLog.objects.filter(
        run_at__lt=datetime.now() - timedelta(days=days)
    )
    logger.info(
        f'Deleting {qs.count} QueryLog objects older than {days} days.'
    )
    qs.delete()
    logger.info('Done deleting QueryLog objects.')


@shared_task
def build_schema_cache_async(connection_alias):
    from .schema import build_schema_info, connection_schema_cache_key
    ret = build_schema_info(connection_alias)
    cache.set(connection_schema_cache_key(connection_alias), ret)
    return ret



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('explorer', '0007_querylog_connection'),
    ]

    operations = [
        migrations.AlterField(
            model_name='query',
            name='connection',
            field=models.CharField(blank=True, help_text='Name of DB connection (as specified in settings) to use for this query. Will use EXPLORER_DEFAULT_CONNECTION if left blank', max_length=128, null=True),
        ),
    ]



# Generated by Django 3.2.16 on 2023-01-12 18:24

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('explorer', '0010_sql_required'),
    ]

    operations = [
        migrations.CreateModel(
            name='QueryFavorite',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('query', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='explorer.query')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('query', 'user')},
            },
        ),
    ]



from django.http import JsonResponse
from django.views.decorators.http import require_POST

from explorer.utils import fmt_sql


@require_POST
def format_sql(request):
    sql = request.POST.get('sql', '')
    formatted = fmt_sql(sql)
    return JsonResponse({"formatted": formatted})



import re
from collections import Counter

from django.forms.models import model_to_dict
from django.views.generic import ListView

from explorer import app_settings
from explorer.models import Query, QueryFavorite, QueryLog
from explorer.utils import allowed_query_pks, url_get_query_id
from explorer.views.auth import PermissionRequiredMixin
from explorer.views.mixins import ExplorerContextMixin


class ListQueryView(PermissionRequiredMixin, ExplorerContextMixin, ListView):
    permission_required = 'view_permission_list'
    model = Query

    def recently_viewed(self):
        qll = QueryLog.objects.filter(
            run_by_user=self.request.user, query_id__isnull=False
        ).order_by(
            '-run_at'
        ).select_related('query')

        ret = []
        tracker = []
        for ql in qll:
            if len(ret) == app_settings.EXPLORER_RECENT_QUERY_COUNT:
                break

            if ql.query_id not in tracker:
                ret.append(ql)
                tracker.append(ql.query_id)
        return ret

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['object_list'] = self._build_queries_and_headers()
        context['recent_queries'] = self.recently_viewed()
        context['tasks_enabled'] = app_settings.ENABLE_TASKS
        return context

    def get_queryset(self):
        if app_settings.EXPLORER_PERMISSION_VIEW(self.request):
            qs = (
                Query.objects.prefetch_related(
                    'created_by_user', 'querylog_set'
                ).all()
            )
        else:
            qs = (
                Query.objects.prefetch_related(
                    'created_by_user', 'querylog_set'
                ).filter(pk__in=allowed_query_pks(self.request.user.id))
            )
        return qs

    def _build_queries_and_headers(self):
        """
        Build a list of query information and headers (pseudo-folders) for
        consumption by the template.

        Strategy: Look for queries with titles of the form "something - else"
                  (eg. with a ' - ' in the middle) and split on the ' - ',
                  treating the left side as a "header" (or folder).
                  Interleave the headers into the ListView's object_list as
                  appropriate. Ignore headers that only have one child.
                  The front end uses bootstrap's JS Collapse plugin, which
                  necessitates generating CSS classes to map the header onto
                  the child rows, hence the collapse_target variable.

                  To make the return object homogeneous, convert the
                  object_list models into dictionaries for interleaving with
                  the header "objects". This necessitates special handling of
                  'created_at' and 'created_by_user' because model_to_dict
                  doesn't include non-editable fields (created_at) and will
                  give the int representation of the user instead of the
                  string representation.

        :return: A list of model dictionaries representing all the query
                 objects, interleaved with header dictionaries.
        :rtype: list
        """

        dict_list = []
        rendered_headers = []
        pattern = re.compile(r'[\W_]+')

        headers = Counter([q.title.split(' - ')[0] for q in self.object_list])
        query_favorites_for_user = QueryFavorite.objects.filter(user_id=self.request.user.id).values_list('query_id',
                                                                                                          flat=True)

        for q in self.object_list:
            model_dict = model_to_dict(q)
            header = q.title.split(' - ')[0]
            collapse_target = pattern.sub('', header)

            if headers[header] > 1 and header not in rendered_headers:
                dict_list.append({
                    'title': header,
                    'is_header': True,
                    'is_in_category': False,
                    'collapse_target': collapse_target,
                    'count': headers[header]
                })
                rendered_headers.append(header)

            model_dict.update({
                'is_in_category': headers[header] > 1,
                'collapse_target': collapse_target,
                'created_at': q.created_at,
                'is_header': False,
                'run_count': q.querylog_set.count(),
                'created_by_user':
                    str(q.created_by_user) if q.created_by_user else None,
                'is_favorite': q.id in query_favorites_for_user
            })
            dict_list.append(model_dict)
        return dict_list


class ListQueryLogView(PermissionRequiredMixin, ExplorerContextMixin, ListView):
    context_object_name = "recent_logs"
    model = QueryLog
    paginate_by = 20
    permission_required = 'view_permission'

    def get_queryset(self):
        kwargs = {'sql__isnull': False}
        if url_get_query_id(self.request):
            kwargs['query_id'] = url_get_query_id(self.request)
        return QueryLog.objects.filter(**kwargs).all()



from django.http import JsonResponse
from django.views import View

from explorer.models import QueryFavorite
from explorer.views.auth import PermissionRequiredMixin
from explorer.views.mixins import ExplorerContextMixin


class QueryFavoritesView(PermissionRequiredMixin, ExplorerContextMixin, View):
    permission_required = 'view_permission'

    def get(self, request):
        favorites = QueryFavorite.objects.filter(user=request.user).select_related('query', 'user').order_by(
            'query__title')
        return self.render_template(
            'explorer/query_favorites.html', {'favorites': favorites}
        )


class QueryFavoriteView(PermissionRequiredMixin, ExplorerContextMixin, View):
    permission_required = 'view_permission'

    @staticmethod
    def build_favorite_response(user, query_id):
        is_favorite = QueryFavorite.objects.filter(user=user, query_id=query_id).exists()
        data = {
            'status': 'success',
            'query_id': query_id,
            'is_favorite': is_favorite
        }
        return data

    def get(self, request, query_id):
        return JsonResponse(QueryFavoriteView.build_favorite_response(request.user, query_id))

    def post(self, request, query_id):
        # toggle favorite
        if QueryFavorite.objects.filter(user=request.user, query_id=query_id).exists():
            QueryFavorite.objects.filter(user=request.user, query_id=query_id).delete()
        else:
            QueryFavorite.objects.get_or_create(user=request.user, query_id=query_id)
        return JsonResponse(QueryFavoriteView.build_favorite_response(request.user, query_id))



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('explorer', '0003_query_snapshot'),
    ]

    operations = [
        migrations.AddField(
            model_name='querylog',
            name='duration',
            field=models.FloatField(null=True, blank=True),
        ),
    ]



from .auth import PermissionRequiredMixin, SafeLoginView
from .create import CreateQueryView
from .delete import DeleteQueryView
from .download import DownloadFromSqlView, DownloadQueryView
from .email import EmailCsvQueryView
from .format_sql import format_sql
from .list import ListQueryLogView, ListQueryView
from .query import PlayQueryView, QueryView
from .query_favorite import QueryFavoritesView, QueryFavoriteView
from .schema import SchemaView
from .stream import StreamQueryView


__all__ = [
    'CreateQueryView',
    'DeleteQueryView',
    'DownloadQueryView',
    'DownloadFromSqlView',
    'EmailCsvQueryView',
    'ListQueryView',
    'ListQueryLogView',
    'PermissionRequiredMixin',
    'PlayQueryView',
    'QueryView',
    'SafeLoginView',
    'StreamQueryView',
    'SchemaView',
    'format_sql',
    'QueryFavoritesView',
    'QueryFavoriteView',
]



import tempfile
from collections import defaultdict
from datetime import date
from wsgiref.util import FileWrapper
from zipfile import ZipFile

from django.http import HttpResponse

from explorer.exporters import CSVExporter


def generate_report_action(description="Generate CSV file from SQL query",):

    def generate_report(modeladmin, request, queryset):
        results = [
            report for report in queryset if report.passes_blacklist()[0]
        ]
        queries = (len(results) > 0 and _package(results)) or defaultdict(int)
        response = HttpResponse(
            queries["data"],
            content_type=queries["content_type"]
        )
        response['Content-Disposition'] = queries["filename"]
        response['Content-Length'] = queries["length"]
        return response

    generate_report.short_description = description
    return generate_report


def _package(queries):
    ret = {}
    is_one = len(queries) == 1
    name_root = lambda n: f"attachment; filename={n}"  # noqa
    ret["content_type"] = (is_one and 'text/csv') or 'application/zip'

    ret["filename"] = (
        is_one and name_root('%s.csv' % queries[0].title.replace(',', ''))
    ) or name_root("Report_%s.zip" % date.today())

    ret["data"] = (
        is_one and CSVExporter(queries[0]).get_output()
    ) or _build_zip(queries)

    ret["length"] = (is_one and len(ret["data"]) or ret["data"].blksize)
    return ret


def _build_zip(queries):
    temp = tempfile.TemporaryFile()
    zip_file = ZipFile(temp, 'w')
    for r in queries:
        zip_file.writestr(
            f'{r.title}.csv', CSVExporter(r).get_output() or "Error!"
        )
    zip_file.close()
    ret = FileWrapper(temp)
    temp.seek(0)
    return ret



from django.shortcuts import get_object_or_404
from django.views.generic.base import View

from explorer.models import Query
from explorer.views.auth import PermissionRequiredMixin
from explorer.views.export import _export


class DownloadQueryView(PermissionRequiredMixin, View):

    permission_required = 'view_permission'

    def get(self, request, query_id, *args, **kwargs):
        query = get_object_or_404(Query, pk=query_id)
        return _export(request, query)


class DownloadFromSqlView(PermissionRequiredMixin, View):

    permission_required = 'view_permission'

    def post(self, request, *args, **kwargs):
        sql = request.POST.get('sql', '')
        connection = request.POST.get('connection', '')
        query = Query(sql=sql, connection=connection, title='')
        ql = query.log(request.user)
        query.title = f'Playground-{ql.id}'
        return _export(request, query)



import unittest
from unittest.mock import patch

from django.core.cache import cache
from django.db import connection
from django.test import TestCase

from explorer import app_settings, schema
from explorer.app_settings import EXPLORER_DEFAULT_CONNECTION as CONN


class TestSchemaInfo(TestCase):

    def setUp(self):
        cache.clear()

    @patch('explorer.schema._get_includes')
    @patch('explorer.schema._get_excludes')
    def test_schema_info_returns_valid_data(self, mocked_excludes,
                                            mocked_includes):
        mocked_includes.return_value = None
        mocked_excludes.return_value = []
        res = schema.schema_info(CONN)
        assert mocked_includes.called  # sanity check: ensure patch worked
        tables = [x[0] for x in res]
        self.assertIn('explorer_query', tables)

    @patch('explorer.schema._get_includes')
    @patch('explorer.schema._get_excludes')
    def test_table_exclusion_list(self, mocked_excludes, mocked_includes):
        mocked_includes.return_value = None
        mocked_excludes.return_value = ('explorer_',)
        res = schema.schema_info(CONN)
        tables = [x[0] for x in res]
        self.assertNotIn('explorer_query', tables)

    @patch('explorer.schema._get_includes')
    @patch('explorer.schema._get_excludes')
    def test_app_inclusion_list(self, mocked_excludes, mocked_includes):
        mocked_includes.return_value = ('auth_',)
        mocked_excludes.return_value = []
        res = schema.schema_info(CONN)
        tables = [x[0] for x in res]
        self.assertNotIn('explorer_query', tables)
        self.assertIn('auth_user', tables)

    @patch('explorer.schema._get_includes')
    @patch('explorer.schema._get_excludes')
    def test_app_inclusion_list_excluded(self, mocked_excludes,
                                         mocked_includes):
        # Inclusion list "wins"
        mocked_includes.return_value = ('explorer_',)
        mocked_excludes.return_value = ('explorer_',)
        res = schema.schema_info(CONN)
        tables = [x[0] for x in res]
        self.assertIn('explorer_query', tables)

    @patch('explorer.schema._include_views')
    def test_app_include_views(self, mocked_include_views):
        database_view = setup_sample_database_view()
        mocked_include_views.return_value = True
        res = schema.schema_info(CONN)
        tables = [x[0] for x in res]
        self.assertIn(database_view, tables)

    @patch('explorer.schema._include_views')
    def test_app_exclude_views(self, mocked_include_views):
        database_view = setup_sample_database_view()
        mocked_include_views.return_value = False
        res = schema.schema_info(CONN)
        tables = [x[0] for x in res]
        self.assertNotIn(database_view, tables)

    @unittest.skipIf(not app_settings.ENABLE_TASKS, 'tasks not enabled')
    @patch('explorer.schema.do_async')
    def test_builds_async(self, mocked_async_check):
        mocked_async_check.return_value = True
        self.assertIsNone(schema.schema_info(CONN))
        res = schema.schema_info(CONN)
        tables = [x[0] for x in res]
        self.assertIn('explorer_query', tables)


def setup_sample_database_view():
    with connection.cursor() as cursor:
        cursor.execute(
            "CREATE VIEW IF NOT EXISTS v_explorer_query AS SELECT title, "
            "sql from explorer_query"
        )
    return 'v_explorer_query'



from django.conf import settings
from django.shortcuts import render

from explorer import app_settings


class ExplorerContextMixin:

    def gen_ctx(self):
        return {
            'can_view': app_settings.EXPLORER_PERMISSION_VIEW(
                self.request
            ),
            'can_change': app_settings.EXPLORER_PERMISSION_CHANGE(
                self.request
            ),
            'csrf_cookie_name': settings.CSRF_COOKIE_NAME,
            'view_name': self.request.resolver_match.view_name,
        }

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx.update(self.gen_ctx())
        return ctx

    def render_template(self, template, ctx):
        ctx.update(self.gen_ctx())
        return render(self.request, template, ctx)



from django.contrib import admin

from explorer.actions import generate_report_action
from explorer.models import Query


@admin.register(Query)
class QueryAdmin(admin.ModelAdmin):
    list_display = ('title', 'description', 'created_by_user',)
    list_filter = ('title',)
    raw_id_fields = ('created_by_user',)
    actions = [generate_report_action()]



import re
from collections import deque
from typing import Iterable, Tuple

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.views import LoginView

import sqlparse
from sqlparse import format as sql_format
from sqlparse.sql import Token, TokenList
from sqlparse.tokens import Keyword

from explorer import app_settings


EXPLORER_PARAM_TOKEN = "$$"


def passes_blacklist(sql: str) -> Tuple[bool, Iterable[str]]:
    sql_strings = sqlparse.split(sql)
    keyword_tokens = set()
    for sql_string in sql_strings:
        statements = sqlparse.parse(sql_string)
        for statement in statements:
            for token in walk_tokens(statement):
                if not token.is_whitespace and not isinstance(token, TokenList):
                    if token.ttype in Keyword:
                        keyword_tokens.add(str(token.value).upper())

    fails = [
        bl_word
        for bl_word in app_settings.EXPLORER_SQL_BLACKLIST
        if bl_word.upper() in keyword_tokens
    ]

    return not bool(fails), fails


def walk_tokens(token: TokenList) -> Iterable[Token]:
    """
    Generator to walk all tokens in a Statement
    https://stackoverflow.com/questions/54982118/parse-case-when-statements-with-sqlparse
    :param token: TokenList
    """
    queue = deque([token])
    while queue:
        token = queue.popleft()
        if isinstance(token, TokenList):
            queue.extend(token)
        yield token


def _format_field(field):
    return field.get_attname_column()[1], field.get_internal_type()


def param(name):
    return f"{EXPLORER_PARAM_TOKEN}{name}{EXPLORER_PARAM_TOKEN}"


def swap_params(sql, params):
    p = params.items() if params else {}
    for k, v in p:
        regex = re.compile(r"\$\$%s(?:\|([^\$\:]+))?(?:\:([^\$]+))?\$\$" % str(k).lower(), re.I)
        sql = regex.sub(str(v), sql)
    return sql


def extract_params(text):
    regex = re.compile(r"\$\$([a-z0-9_]+)(?:\|([^\$\:]+))?(?:\:([^\$]+))?\$\$", re.IGNORECASE)
    params = re.findall(regex, text)
    # Matching will result to ('name', 'label', 'default')
    return {
        p[0].lower(): {
            'label': p[1],
            'default': p[2]
        } for p in params if len(p) > 1
    }


def safe_login_prompt(request):
    defaults = {
        'template_name': 'admin/login.html',
        'authentication_form': AuthenticationForm,
        'extra_context': {
            'title': 'Log in',
            'app_path': request.get_full_path(),
            REDIRECT_FIELD_NAME: request.get_full_path(),
        },
    }
    return LoginView.as_view(**defaults)(request)


def shared_dict_update(target, source):
    for k_d1 in target:
        if k_d1 in source:
            target[k_d1] = source[k_d1]
    return target


def safe_cast(val, to_type, default=None):
    try:
        return to_type(val)
    except ValueError:
        return default


def get_int_from_request(request, name, default):
    val = request.GET.get(name, default)
    return safe_cast(val, int, default) if val else None


def get_params_from_request(request):
    val = request.GET.get('params', None)
    try:
        d = {}
        tuples = val.split('|')
        for t in tuples:
            res = t.split(':')
            d[res[0]] = res[1]
        return d
    except Exception:
        return None


def get_params_for_url(query):
    if query.params:
        return '|'.join([f'{p}:{v}' for p, v in query.params.items()])


def url_get_rows(request):
    return get_int_from_request(
        request, 'rows', app_settings.EXPLORER_DEFAULT_ROWS
    )


def url_get_query_id(request):
    return get_int_from_request(request, 'query_id', None)


def url_get_log_id(request):
    return get_int_from_request(request, 'querylog_id', None)


def url_get_show(request):
    return bool(get_int_from_request(request, 'show', 1))


def url_get_fullscreen(request):
    return bool(get_int_from_request(request, 'fullscreen', 0))


def url_get_params(request):
    return get_params_from_request(request)


def allowed_query_pks(user_id):
    return app_settings.EXPLORER_GET_USER_QUERY_VIEWS().get(user_id, [])


def user_can_see_query(request, **kwargs):
    if not request.user.is_anonymous and 'query_id' in kwargs:
        return int(kwargs['query_id']) in allowed_query_pks(request.user.id)
    return False


def fmt_sql(sql):
    return sql_format(sql, reindent=True, keyword_case='upper')


def noop_decorator(f):
    return f


class InvalidExplorerConnectionException(Exception):
    pass


def get_valid_connection(alias=None):
    from explorer.connections import connections

    if not alias:
        return connections[app_settings.EXPLORER_DEFAULT_CONNECTION]

    if alias not in connections:
        raise InvalidExplorerConnectionException(
            f'Attempted to access connection {alias}, '
            f'but that is not a registered Explorer connection.'
        )
    return connections[alias]


def get_s3_bucket():
    import boto3
    kwargs = {
        'aws_access_key_id': app_settings.S3_ACCESS_KEY,
        'aws_secret_access_key': app_settings.S3_SECRET_KEY,
        'region_name': app_settings.S3_REGION
    }
    if app_settings.S3_ENDPOINT_URL:
        kwargs['endpoint_url'] = app_settings.S3_ENDPOINT_URL
    s3 = boto3.resource('s3', **kwargs)
    return s3.Bucket(name=app_settings.S3_BUCKET)


def s3_upload(key, data):
    bucket = get_s3_bucket()
    bucket.upload_fileobj(data, key, ExtraArgs={'ContentType': "text/csv"})
    return s3_url(bucket, key)


def s3_url(bucket, key):
    url = bucket.meta.client.generate_presigned_url(
        ClientMethod='get_object',
        Params={'Bucket': app_settings.S3_BUCKET, 'Key': key},
        ExpiresIn=app_settings.S3_LINK_EXPIRATION)
    return url


def is_xls_writer_available():
    try:
        import xlsxwriter  # noqa
        return True
    except ImportError:
        return False



from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import path

from explorer.urls import urlpatterns


admin.autodiscover()

urlpatterns += [
    path('admin/', admin.site.urls),
]

urlpatterns += staticfiles_urlpatterns()



from django.db.utils import IntegrityError
from django.forms.models import model_to_dict
from django.test import TestCase

from explorer.forms import QueryForm
from explorer.tests.factories import SimpleQueryFactory


class TestFormValidation(TestCase):

    def test_form_is_valid_with_valid_sql(self):
        q = SimpleQueryFactory(sql="select 1;", created_by_user_id=None)
        form = QueryForm(model_to_dict(q))
        self.assertTrue(form.is_valid())

    def test_form_fails_null(self):
        with self.assertRaises(IntegrityError):
            SimpleQueryFactory(sql=None, created_by_user_id=None)

    def test_form_fails_blank(self):
        q = SimpleQueryFactory(sql='', created_by_user_id=None)
        q.params = {}
        form = QueryForm(model_to_dict(q))
        self.assertFalse(form.is_valid())

    def test_form_fails_blacklist(self):
        q = SimpleQueryFactory(sql="delete $$a$$;", created_by_user_id=None)
        q.params = {}
        form = QueryForm(model_to_dict(q))
        self.assertFalse(form.is_valid())



from django.apps import AppConfig
from django.core.exceptions import ImproperlyConfigured
from django.db import connections as djcs
from django.utils.translation import gettext_lazy as _


class ExplorerAppConfig(AppConfig):

    name = 'explorer'
    verbose_name = _('SQL Explorer')
    default_auto_field = 'django.db.models.AutoField'

    def ready(self):
        from explorer.schema import build_async_schemas
        _validate_connections()
        build_async_schemas()


def _get_default():
    from explorer.app_settings import EXPLORER_DEFAULT_CONNECTION
    return EXPLORER_DEFAULT_CONNECTION


def _get_explorer_connections():
    from explorer.app_settings import EXPLORER_CONNECTIONS
    return EXPLORER_CONNECTIONS


def _validate_connections():

    # Validate connections
    if _get_default() not in _get_explorer_connections().values():
        raise ImproperlyConfigured(
            f'EXPLORER_DEFAULT_CONNECTION is {_get_default()}, '
            f'but that alias is not present in the values of '
            f'EXPLORER_CONNECTIONS'
        )

    for name, conn_name in _get_explorer_connections().items():
        if conn_name not in djcs:
            raise ImproperlyConfigured(
                f'EXPLORER_CONNECTIONS contains ({name}, {conn_name}), '
                f'but {conn_name} is not a valid Django DB connection.'
            )



import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Query',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('title', models.CharField(max_length=255)),
                ('sql', models.TextField(blank=True)),
                ('description', models.TextField(blank=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('last_run_date', models.DateTimeField(auto_now=True)),
                ('created_by_user', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, null=True, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['title'],
                'verbose_name_plural': 'Queries',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='QueryLog',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('sql', models.TextField(blank=True)),
                ('is_playground', models.BooleanField(default=False)),
                ('run_at', models.DateTimeField(auto_now_add=True)),
                ('query', models.ForeignKey(on_delete=django.db.models.deletion.SET_NULL, blank=True, to='explorer.Query', null=True)),
                ('run_by_user', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, null=True, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['-run_at'],
            },
            bases=(models.Model,),
        ),
    ]



from io import BytesIO
from typing import Iterable, Optional

from django.core.exceptions import ImproperlyConfigured

from explorer import app_settings


if app_settings.EXPLORER_CHARTS_ENABLED:
    try:
        import matplotlib.pyplot as plt
        import seaborn as sns
        from matplotlib.figure import Figure
    except ImportError:
        raise ImproperlyConfigured(
            "If `EXPLORER_CHARTS_ENABLED` is enabled, `matplotlib` and `seaborn` must be installed.")

from .models import QueryResult


def get_pie_chart(result: QueryResult) -> Optional[str]:
    """
    Return a pie chart in SVG format if the result table adheres to the expected format.

    A pie chart is rendered if
    * there is at least on row of in the result table
    * the result table has at least two columns
    * the second column is of a numeric type

    The first column is used as labels for the pie sectors.
    The second column is used to determine the size of the sectors
    (hence the requirement of it being numeric).
    All other columns are ignored.
    """
    if len(result.data) < 1 or len(result.data[0]) < 2:
        return None
    not_none_rows = [row for row in result.data if row[0] is not None and row[1] is not None]
    labels = [row[0] for row in not_none_rows]
    values = [row[1] for row in not_none_rows]
    if not is_numeric(values):
        return None
    fig, ax = plt.subplots(figsize=(4.5, 4.5))
    ax.pie(values, labels=labels)
    return get_svg(fig)


def get_line_chart(result: QueryResult) -> Optional[str]:
    """
    Return a line chart in SVG format if the result table adheres to the expected format.

    A line chart is rendered if
    * there is at least on row of in the result table
    * there is at least one numeric column (the first column (with index 0) does not count)

    The first column is used as x-axis labels.
    All other numeric columns represent a line on the chart.
    The name of the column is used as the name of the line in the legend.
    Not numeric columns (except the first on) are ignored.
    """
    if len(result.data) < 1:
        return None
    numeric_columns = [c for c in range(1, len(result.data[0]))
                       if all([isinstance(col[c], (int, float)) or col[c] is None for col in result.data])]
    if len(numeric_columns) < 1:
        return None
    labels = [row[0] for row in result.data]
    fig, ax = plt.subplots(figsize=(10, 3.8))
    for col_num in numeric_columns:
        sns.lineplot(ax=ax,
                     x=labels,
                     y=[row[col_num] for row in result.data],
                     label=result.headers[col_num])
    ax.set_xlabel(result.headers[0])
    # Rotate x-axis labels by 20 degrees to reduce overlap
    for label in ax.get_xticklabels():
        label.set_rotation(20)
        label.set_ha('right')
    return get_svg(fig)


def get_svg(fig: 'Figure') -> str:
    buffer = BytesIO()
    fig.savefig(buffer, format='svg')
    buffer.seek(0)
    graph = buffer.getvalue().decode('utf-8')
    buffer.close()
    return graph


def is_numeric(column: Iterable) -> bool:
    """
    Indicate if all the values in the given column are numeric or None.
    """
    return all([isinstance(value, (int, float)) or value is None for value in column])



from django import template
from django.utils.module_loading import import_string

from explorer import app_settings


register = template.Library()


@register.inclusion_tag('explorer/export_buttons.html')
def export_buttons(query=None):
    exporters = []
    for name, classname in app_settings.EXPLORER_DATA_EXPORTERS:
        exporter_class = import_string(classname)
        exporters.append((name, exporter_class.name))
    return {
        'exporters': exporters,
        'query': query,
    }


@register.inclusion_tag('explorer/query_favorite_button.html')
def query_favorite_button(query_id, is_favorite, extra_classes):
    return {
        'query_id': query_id,
        'is_favorite': is_favorite,
        'extra_classes': extra_classes
    }



# Generated by Django 3.1.2 on 2020-10-09 05:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('explorer', '0008_auto_20190308_1642'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='query',
            options={'ordering': ['title'], 'verbose_name': 'Query', 'verbose_name_plural': 'Queries'},
        ),
        migrations.AlterField(
            model_name='query',
            name='connection',
            field=models.CharField(blank=True, default='', help_text='Name of DB connection (as specified in settings) to use for this query.Will use EXPLORER_DEFAULT_CONNECTION if left blank', max_length=128),
        ),
        migrations.AlterField(
            model_name='querylog',
            name='connection',
            field=models.CharField(blank=True, default='', max_length=128),
        ),
    ]



# Generated by Django 3.0.8 on 2020-12-18 06:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('explorer', '0009_auto_20201009_0547'),
    ]

    operations = [
        migrations.AlterField(
            model_name='query',
            name='sql',
            field=models.TextField(),
        ),
    ]



from pydoc import locate

from django.conf import settings


# The 'correct' configuration for Explorer looks like:

# EXPLORER_CONNECTIONS = {
#   'Original Database': 'my_important_database_readonly_connection',
#   'Client Database 2': 'other_database_connection'
# }
# EXPLORER_DEFAULT_CONNECTION = 'my_important_database_readonly_connection'

EXPLORER_CONNECTIONS = getattr(settings, 'EXPLORER_CONNECTIONS', {})
EXPLORER_DEFAULT_CONNECTION = getattr(
    settings, 'EXPLORER_DEFAULT_CONNECTION', None
)

# Change the behavior of explorer
EXPLORER_SQL_BLACKLIST = getattr(
    settings, 'EXPLORER_SQL_BLACKLIST',
    (
        # DML
        'COMMIT',
        'DELETE',
        'INSERT',
        'MERGE',
        'REPLACE',
        'ROLLBACK',
        'SET',
        'START',
        'UPDATE',
        'UPSERT',

        # DDL
        'ALTER',
        'CREATE',
        'DROP',
        'RENAME',
        'TRUNCATE',

        # DCL
        'GRANT',
        'REVOKE',
    )
)


EXPLORER_DEFAULT_ROWS = getattr(settings, 'EXPLORER_DEFAULT_ROWS', 1000)

EXPLORER_SCHEMA_EXCLUDE_TABLE_PREFIXES = getattr(
    settings,
    'EXPLORER_SCHEMA_EXCLUDE_TABLE_PREFIXES',
    (
        'auth_',
        'contenttypes_',
        'sessions_',
        'admin_'
    )
)

EXPLORER_SCHEMA_INCLUDE_TABLE_PREFIXES = getattr(
    settings,
    'EXPLORER_SCHEMA_INCLUDE_TABLE_PREFIXES',
    None
)
EXPLORER_SCHEMA_INCLUDE_VIEWS = getattr(
    settings,
    'EXPLORER_SCHEMA_INCLUDE_VIEWS',
    False
)

EXPLORER_TRANSFORMS = getattr(settings, 'EXPLORER_TRANSFORMS', [])
EXPLORER_PERMISSION_VIEW = getattr(
    settings, 'EXPLORER_PERMISSION_VIEW', lambda r: r.user.is_staff
)
EXPLORER_PERMISSION_CHANGE = getattr(
    settings, 'EXPLORER_PERMISSION_CHANGE', lambda r: r.user.is_staff
)
EXPLORER_RECENT_QUERY_COUNT = getattr(
    settings, 'EXPLORER_RECENT_QUERY_COUNT', 10
)
EXPLORER_ASYNC_SCHEMA = getattr(settings, 'EXPLORER_ASYNC_SCHEMA', False)

DEFAULT_EXPORTERS = [
    ('csv', 'explorer.exporters.CSVExporter'),
    ('json', 'explorer.exporters.JSONExporter'),
]
try:
    import xlsxwriter  # noqa

    DEFAULT_EXPORTERS.insert(
        1,
        ('excel', 'explorer.exporters.ExcelExporter'),
    )
except ImportError:
    pass

EXPLORER_DATA_EXPORTERS = getattr(
    settings, 'EXPLORER_DATA_EXPORTERS', DEFAULT_EXPORTERS
)
CSV_DELIMETER = getattr(settings, "EXPLORER_CSV_DELIMETER", ",")

# API access
EXPLORER_TOKEN = getattr(settings, 'EXPLORER_TOKEN', 'CHANGEME')

# These are callable to aid testability by dodging the settings cache.
# There is surely a better pattern for this, but this'll hold for now.
EXPLORER_GET_USER_QUERY_VIEWS = lambda: getattr(  # noqa
    settings, 'EXPLORER_USER_QUERY_VIEWS', {}
)
EXPLORER_TOKEN_AUTH_ENABLED = lambda: getattr(  # noqa
    settings, 'EXPLORER_TOKEN_AUTH_ENABLED', False
)
EXPLORER_NO_PERMISSION_VIEW = lambda: locate(# noqa
    getattr(
        settings,
        'EXPLORER_NO_PERMISSION_VIEW',
        'explorer.views.auth.safe_login_view_wrapper',
    ),
)

# Async task related. Note that the EMAIL_HOST settings must be set up for
# email to work.
ENABLE_TASKS = getattr(settings, "EXPLORER_TASKS_ENABLED", False)
S3_ACCESS_KEY = getattr(settings, "EXPLORER_S3_ACCESS_KEY", None)
S3_SECRET_KEY = getattr(settings, "EXPLORER_S3_SECRET_KEY", None)
S3_BUCKET = getattr(settings, "EXPLORER_S3_BUCKET", None)
S3_LINK_EXPIRATION: int = getattr(settings, "EXPLORER_S3_LINK_EXPIRATION", 3600)
FROM_EMAIL = getattr(
    settings, 'EXPLORER_FROM_EMAIL', 'django-sql-explorer@example.com'
)
S3_REGION = getattr(settings, "EXPLORER_S3_REGION", "us-east-1")
S3_ENDPOINT_URL = getattr(settings, "EXPLORER_S3_ENDPOINT_URL", None)

UNSAFE_RENDERING = getattr(settings, "EXPLORER_UNSAFE_RENDERING", False)

EXPLORER_CHARTS_ENABLED = getattr(settings, "EXPLORER_CHARTS_ENABLED", False)

# If set to True will autorun queries when viewed which is the historical behavior
# Default to True if not set in order to be backwards compatible
# If set to False will not autorun queries containing parameters when viewed
# - user will need to run by clicking the Save & Run Button to execute
EXPLORER_AUTORUN_QUERY_WITH_PARAMS = getattr(settings, "EXPLORER_AUTORUN_QUERY_WITH_PARAMS", True)



import unittest
from datetime import datetime, timedelta
from io import StringIO
from unittest.mock import patch

from django.core import mail
from django.test import TestCase

from explorer import app_settings
from explorer.app_settings import EXPLORER_DEFAULT_CONNECTION as CONN
from explorer.models import QueryLog
from explorer.tasks import build_schema_cache_async, execute_query, snapshot_queries, truncate_querylogs
from explorer.tests.factories import SimpleQueryFactory


class TestTasks(TestCase):

    @unittest.skipIf(not app_settings.ENABLE_TASKS, 'tasks not enabled')
    @patch('explorer.tasks.s3_upload')
    def test_async_results(self, mocked_upload):
        mocked_upload.return_value = 'http://s3.com/your-file.csv'

        q = SimpleQueryFactory(
            sql='select 1 "a", 2 "b", 3 "c";', title="testquery"
        )
        execute_query(q.id, 'cc@epantry.com')

        output = StringIO()
        output.write('a,b,c\r\n1,2,3\r\n')

        self.assertEqual(len(mail.outbox), 2)
        self.assertIn(
            '[SQL Explorer] Your query is running', mail.outbox[0].subject
        )
        self.assertIn('[SQL Explorer] Report ', mail.outbox[1].subject)
        self.assertEqual(
            mocked_upload
            .call_args[0][1].getvalue()
            .decode('utf-8-sig'),
            output.getvalue()
        )
        self.assertEqual(mocked_upload.call_count, 1)

    @unittest.skipIf(not app_settings.ENABLE_TASKS, 'tasks not enabled')
    @patch('explorer.tasks.s3_upload')
    def test_async_results_fails_with_message(self, mocked_upload):
        mocked_upload.return_value = 'http://s3.com/your-file.csv'

        q = SimpleQueryFactory(sql='select x from foo;', title="testquery")
        execute_query(q.id, 'cc@epantry.com')

        output = StringIO()
        output.write('a,b,c\r\n1,2,3\r\n')

        self.assertEqual(len(mail.outbox), 2)
        self.assertIn('[SQL Explorer] Error ', mail.outbox[1].subject)
        self.assertEqual(mocked_upload.call_count, 0)

    @unittest.skipIf(not app_settings.ENABLE_TASKS, 'tasks not enabled')
    @patch('explorer.tasks.s3_upload')
    def test_snapshots(self, mocked_upload):
        mocked_upload.return_value = 'http://s3.com/your-file.csv'

        SimpleQueryFactory(snapshot=True)
        SimpleQueryFactory(snapshot=True)
        SimpleQueryFactory(snapshot=True)
        SimpleQueryFactory(snapshot=False)

        snapshot_queries()
        self.assertEqual(mocked_upload.call_count, 3)

    @unittest.skipIf(not app_settings.ENABLE_TASKS, 'tasks not enabled')
    def test_truncating_querylogs(self):
        QueryLog(sql='foo').save()
        QueryLog.objects.filter(sql='foo').update(
            run_at=datetime.now() - timedelta(days=30)
        )
        QueryLog(sql='bar').save()
        QueryLog.objects.filter(sql='bar').update(
            run_at=datetime.now() - timedelta(days=29)
        )
        truncate_querylogs(30)
        self.assertEqual(QueryLog.objects.count(), 1)

    @unittest.skipIf(not app_settings.ENABLE_TASKS, 'tasks not enabled')
    @patch('explorer.schema.build_schema_info')
    def test_build_schema_cache_async(self, mocked_build):
        mocked_build.return_value = ['list_of_tuples']
        schema = build_schema_cache_async(CONN)
        assert mocked_build.called
        self.assertEqual(schema, ['list_of_tuples'])



__version_info__ = {
    'major': 3,
    'minor': 1,
    'patch': 0,
    'releaselevel': 'final',
    'serial': 0
}


def get_version(short=False):
    assert __version_info__['releaselevel'] in ('alpha', 'beta', 'final')
    vers = ["%(major)i.%(minor)i" % __version_info__, ]
    if __version_info__['patch']:
        vers.append(".%(patch)i" % __version_info__)
    if __version_info__['releaselevel'] != 'final' and not short:
        vers.append(
            '%s%i' % (
                __version_info__['releaselevel'][0],
                __version_info__['serial'])
        )
    return ''.join(vers)


__version__ = get_version()



import json
import unittest
from datetime import date, datetime

from django.core.serializers.json import DjangoJSONEncoder
from django.db import connections
from django.test import TestCase
from django.utils import timezone

from explorer.app_settings import EXPLORER_DEFAULT_CONNECTION as CONN
from explorer.exporters import CSVExporter, ExcelExporter, JSONExporter
from explorer.models import QueryResult
from explorer.tests.factories import SimpleQueryFactory
from explorer.utils import is_xls_writer_available


class TestCsv(TestCase):

    def test_writing_unicode(self):
        res = QueryResult(
            SimpleQueryFactory(sql='select 1 as "a", 2 as ""').sql,
            connections[CONN]
        )
        res.execute_query()
        res.process()
        res._data = [[1, None], ["Jenét", '1']]

        res = CSVExporter(query=None)._get_output(res).getvalue()
        self.assertEqual(
            res.encode('utf-8').decode('utf-8-sig'),
            'a,\r\n1,\r\nJenét,1\r\n'
        )

    def test_custom_delimiter(self):
        q = SimpleQueryFactory(sql='select 1, 2')
        exporter = CSVExporter(query=q)
        res = exporter.get_output(delim='|')
        self.assertEqual(
            res.encode('utf-8').decode('utf-8-sig'),
            '1|2\r\n1|2\r\n'
        )

    def test_writing_bom(self):
        q = SimpleQueryFactory(sql='select 1, 2')
        exporter = CSVExporter(query=q)
        res = exporter.get_output()
        self.assertEqual(res, '\ufeff1,2\r\n1,2\r\n')


class TestJson(TestCase):

    def test_writing_json(self):
        res = QueryResult(
            SimpleQueryFactory(sql='select 1 as "a", 2 as ""').sql,
            connections[CONN]
        )
        res.execute_query()
        res.process()
        res._data = [[1, None], ["Jenét", '1']]

        res = JSONExporter(query=None)._get_output(res).getvalue()
        expected = [{'a': 1, '': None}, {'a': 'Jenét', '': '1'}]
        self.assertEqual(res, json.dumps(expected))

    def test_writing_datetimes(self):
        res = QueryResult(
            SimpleQueryFactory(sql='select 1 as "a", 2 as "b"').sql,
            connections[CONN]
        )
        res.execute_query()
        res.process()
        res._data = [[1, date.today()]]

        res = JSONExporter(query=None)._get_output(res).getvalue()
        expected = [{'a': 1, 'b': date.today()}]
        self.assertEqual(res, json.dumps(expected, cls=DjangoJSONEncoder))


class TestExcel(TestCase):

    @unittest.skipIf(not is_xls_writer_available(), 'excel exporter not available')
    def test_writing_excel(self):
        """
        This is a pretty crap test. It at least exercises the code.
        If anyone wants to go through the brain damage of actually building
        an 'expected' xlsx output and comparing it
        (https://github.com/jmcnamara/XlsxWriter/blob/master/xlsxwriter/
        test/helperfunctions.py)
        by all means submit a pull request!
        """
        res = QueryResult(
            SimpleQueryFactory(
                sql='select 1 as "a", 2 as ""',
                title='\\/*[]:?this title is longer than 32 characters'
            ).sql,
            connections[CONN]
        )

        res.execute_query()
        res.process()

        d = datetime.now()
        d = timezone.make_aware(d, timezone.get_current_timezone())

        res._data = [[1, None], ["Jenét", d]]

        res = ExcelExporter(
            query=SimpleQueryFactory()
        )._get_output(res).getvalue()

        expected = b'PK'

        self.assertEqual(res[:2], expected)

    @unittest.skipIf(not is_xls_writer_available(), 'excel exporter not available')
    def test_writing_dict_fields(self):
        res = QueryResult(
            SimpleQueryFactory(
                sql='select 1 as "a", 2 as ""',
                title='\\/*[]:?this title is longer than 32 characters'
            ).sql,
            connections[CONN]
        )

        res.execute_query()
        res.process()

        res._data = [[1, ['foo', 'bar']], [2, {'foo': 'bar'}]]

        res = ExcelExporter(
            query=SimpleQueryFactory()
        )._get_output(res).getvalue()

        expected = b'PK'

        self.assertEqual(res[:2], expected)



from unittest.mock import Mock

from django.test import TestCase

from explorer import app_settings
from explorer.tests.factories import SimpleQueryFactory
from explorer.utils import (
    EXPLORER_PARAM_TOKEN, extract_params, get_params_for_url, get_params_from_request, param, passes_blacklist,
    shared_dict_update, swap_params,
)


class TestSqlBlacklist(TestCase):

    def setUp(self):
        self.orig = app_settings.EXPLORER_SQL_BLACKLIST

    def tearDown(self):
        app_settings.EXPLORER_SQL_BLACKLIST = self.orig

    def test_overriding_blacklist(self):
        app_settings.EXPLORER_SQL_BLACKLIST = []
        sql = "DELETE FROM some_table;"
        passes, words = passes_blacklist(sql)
        self.assertTrue(passes)

    def test_not_overriding_blacklist(self):
        sql = "DELETE FROM some_table;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    # Various flavors of select - all should be ok
    def test_select_keywords_as_literals(self):
        sql = "SELECT * from eventtype where eventtype.value = 'Grant Date';"
        passes, words = passes_blacklist(sql)
        self.assertTrue(passes)

    def test_select_containing_drop_in_word(self):
        sql = "SELECT * FROM student droptable WHERE name LIKE 'Robert%'"
        self.assertTrue(passes_blacklist(sql)[0])

    def test_select_with_case(self):
        sql = '''SELECT   ProductNumber, Name, "Price Range" =
          CASE
             WHEN ListPrice =  0 THEN 'Mfg item - not for resale'
             WHEN ListPrice < 50 THEN 'Under $50'
             WHEN ListPrice >= 50 and ListPrice < 250 THEN 'Under $250'
             WHEN ListPrice >= 250 and ListPrice < 1000 THEN 'Under $1000'
             ELSE 'Over $1000'
          END
        FROM Production.Product
        ORDER BY ProductNumber ;
        '''
        passes, words = passes_blacklist(sql)
        self.assertTrue(passes)

    def test_select_with_subselect(self):
        sql = '''SELECT a.studentid, a.name, b.total_marks
            FROM student a, marks b
            WHERE a.studentid = b.studentid AND b.total_marks >
            (SELECT total_marks
            FROM marks
            WHERE studentid =  'V002');
            '''
        passes, words = passes_blacklist(sql)
        self.assertTrue(passes)

    def test_select_with_replace_function(self):
        sql = "SELECT replace('test string', 'st', '**');"
        passes, words = passes_blacklist(sql)
        self.assertTrue(passes)

    def test_dml_commit(self):
        sql = "COMMIT TRANSACTION;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dml_delete(self):
        sql = "'distraction'; deLeTe from table; " \
              "SELECT 1+1 AS TWO; drop view foo;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)
        self.assertEqual(len(words), 2)

    def test_dml_insert(self):
        sql = "INSERT INTO products (product_no, name, price) VALUES (1, 'Cheese', 9.99);"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dml_merge(self):
        sql = '''MERGE INTO wines w
            USING (VALUES('Chateau Lafite 2003', '24')) v
            ON v.column1 = w.winename
            WHEN NOT MATCHED
              INSERT VALUES(v.column1, v.column2)
            WHEN MATCHED
              UPDATE SET stock = stock + v.column2;'''
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dml_replace(self):
        sql = "REPLACE INTO test VALUES (1, 'Old', '2014-08-20 18:47:00');"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dml_rollback(self):
        sql = "ROLLBACK TO SAVEPOINT my_savepoint;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dml_set(self):
        sql = "SET TIME ZONE 'PST8PDT';"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dml_start(self):
        sql = "START TRANSACTION;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dml_update(self):
        sql = '''UPDATE accounts SET (contact_first_name, contact_last_name) =
        (SELECT first_name, last_name FROM employees
         WHERE employees.id = accounts.sales_person);'''
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dml_upsert(self):
        sql = "UPSERT INTO Users VALUES (10, 'John', 'Smith', 27, 60000);"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_ddl_alter(self):
        sql = '''ALTER TABLE foo
        ALTER COLUMN foo_timestamp DROP DEFAULT,
        ALTER COLUMN foo_timestamp TYPE timestamp with time zone
        USING
            timestamp with time zone 'epoch' + foo_timestamp * interval '1 second',
        ALTER COLUMN foo_timestamp SET DEFAULT now();'''
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_ddl_create(self):
        sql = '''CREATE TABLE Persons (
            PersonID int,
            LastName varchar(255),
            FirstName varchar(255),
            Address varchar(255),
            City varchar(255)
        );
        '''
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_ddl_drop(self):
        sql = "DROP TABLE films, distributors;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_ddl_rename(self):
        sql = "RENAME TABLE old_table_name TO new_table_name;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_ddl_truncate(self):
        sql = "TRUNCATE bigtable, othertable RESTART IDENTITY;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dcl_grant(self):
        sql = "GRANT ALL PRIVILEGES ON kinds TO manuel;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dcl_revoke(self):
        sql = "REVOKE ALL PRIVILEGES ON kinds FROM manuel;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)

    def test_dcl_revoke_bad_syntax(self):
        sql = "REVOKE ON kinds; FROM manuel;"
        passes, words = passes_blacklist(sql)
        self.assertFalse(passes)


class TestParams(TestCase):

    def test_swappable_params_are_built_correctly(self):
        expected = EXPLORER_PARAM_TOKEN + 'foo' + EXPLORER_PARAM_TOKEN
        self.assertEqual(expected, param('foo'))

    def test_params_get_swapped(self):
        sql = 'please Swap $$this$$ and $$THat$$'
        expected = 'please Swap here and there'
        params = {'this': 'here', 'that': 'there'}
        got = swap_params(sql, params)
        self.assertEqual(got, expected)

    def test_empty_params_does_nothing(self):
        sql = 'please swap $$this$$ and $$that$$'
        params = None
        got = swap_params(sql, params)
        self.assertEqual(got, sql)

    def test_non_string_param_gets_swapper(self):
        sql = 'please swap $$this$$'
        expected = 'please swap 1'
        params = {'this': 1}
        got = swap_params(sql, params)
        self.assertEqual(got, expected)

    def _assertSwap(self, tuple):
        self.assertEqual(extract_params(tuple[0]), tuple[1])

    def test_extracting_params(self):
        tests = [
            ('please swap $$this0$$', {'this0': {'default': '', 'label': ''}}),
            ('please swap $$THis0$$', {'this0': {'default': '', 'label': ''}}),
            ('please swap $$this6$$ $$this6:that$$', {'this6': {'default': 'that', 'label': ''}}),
            ('please swap $$this_7:foo, bar$$', {'this_7': {'default': 'foo, bar', 'label': ''}}),
            ('please swap $$this8:$$', {}),
            ('do nothing with $$this1 $$', {}),
            ('do nothing with $$this2 :$$', {}),
            ('do something with $$this3: $$', {'this3': {'default': ' ', 'label': ''}}),
            ('do nothing with $$this4: ', {}),
            ('do nothing with $$this5$that$$', {}),
            ('check label $$this|label:val$$', {'this': {'default': 'val', 'label': 'label'}}),
            ('check case $$this|label Case:Va l$$', {'this': {'default': 'Va l', 'label': 'label Case'}}),
            ('check label case and unicode $$this|label Case ελληνικά:val Τέστ$$', {
                'this': {'default': 'val Τέστ', 'label': 'label Case ελληνικά'}
            }),
        ]
        for s in tests:
            self._assertSwap(s)

    def test_shared_dict_update(self):
        source = {'foo': 1, 'bar': 2}
        target = {'bar': None}  # ha ha!
        self.assertEqual({'bar': 2}, shared_dict_update(target, source))

    def test_get_params_from_url(self):
        r = Mock()
        r.GET = {'params': 'foo:bar|qux:mux'}
        res = get_params_from_request(r)
        self.assertEqual(res['foo'], 'bar')
        self.assertEqual(res['qux'], 'mux')

    def test_get_params_for_request(self):
        q = SimpleQueryFactory(params={'a': 1, 'b': 2})
        # For some reason the order of the params is non-deterministic,
        # causing the following to periodically fail:
        #     self.assertEqual(get_params_for_url(q), 'a:1|b:2')
        # So instead we go for the following, convoluted, asserts:
        res = get_params_for_url(q)
        res = res.split('|')
        expected = ['a:1', 'b:2']
        for e in expected:
            self.assertIn(e, res)

    def test_get_params_for_request_empty(self):
        q = SimpleQueryFactory()
        self.assertEqual(get_params_for_url(q), None)


class TestConnections(TestCase):

    def test_only_registered_connections_are_in_connections(self):
        from django.db import connections as djcs

        from explorer.app_settings import EXPLORER_DEFAULT_CONNECTION
        from explorer.connections import connections
        self.assertTrue(EXPLORER_DEFAULT_CONNECTION in connections)
        self.assertNotEqual(len(connections), len([c for c in djcs]))



import importlib
import json
import time
import unittest
from unittest.mock import Mock, patch

from django.contrib.auth.models import User
from django.core.cache import cache
from django.db import connections
from django.forms.models import model_to_dict
from django.shortcuts import redirect
from django.test import TestCase
from django.urls import reverse

from explorer import app_settings
from explorer.app_settings import EXPLORER_DEFAULT_CONNECTION as CONN
from explorer.app_settings import EXPLORER_TOKEN
from explorer.models import MSG_FAILED_BLACKLIST, Query, QueryFavorite, QueryLog
from explorer.tests.factories import QueryLogFactory, SimpleQueryFactory
from explorer.utils import user_can_see_query


def reload_app_settings():
    """
    Reload app settings, otherwise changes from testing context manager won't take effect
    app_settings are loaded at time of import
    """
    importlib.reload(app_settings)


class TestQueryListView(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    def test_admin_required(self):
        self.client.logout()
        resp = self.client.get(reverse("explorer_index"))
        self.assertTemplateUsed(resp, 'admin/login.html')

    def test_headers(self):
        SimpleQueryFactory(title='foo - bar1')
        SimpleQueryFactory(title='foo - bar2')
        SimpleQueryFactory(title='foo - bar3')
        SimpleQueryFactory(title='qux - mux')
        resp = self.client.get(reverse("explorer_index"))
        self.assertContains(resp, 'foo (3)')
        self.assertContains(resp, 'foo - bar2')
        self.assertContains(resp, 'qux - mux')

    def test_permissions_show_only_allowed_queries(self):
        self.client.logout()
        q1 = SimpleQueryFactory(title='canseethisone')
        q2 = SimpleQueryFactory(title='nope')
        user = User.objects.create_user('user', 'user@user.com', 'pwd')
        self.client.login(username='user', password='pwd')
        with self.settings(EXPLORER_USER_QUERY_VIEWS={user.id: [q1.id]}):
            resp = self.client.get(reverse("explorer_index"))
        self.assertTemplateUsed(resp, 'explorer/query_list.html')
        self.assertContains(resp, q1.title)
        self.assertNotContains(resp, q2.title)

    def test_run_count(self):
        q = SimpleQueryFactory(title='foo - bar1')
        for i in range(0, 4):
            q.log()
        resp = self.client.get(reverse("explorer_index"))
        self.assertContains(resp, '<td>4</td>')


class TestQueryCreateView(TestCase):

    def setUp(self):
        self.admin = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.user = User.objects.create_user(
            'user', 'user@user.com', 'pwd'
        )

    def test_change_permission_required(self):
        self.client.login(username='user', password='pwd')
        resp = self.client.get(reverse("query_create"))
        self.assertTemplateUsed(resp, 'admin/login.html')

    def test_renders_with_title(self):
        self.client.login(username='admin', password='pwd')
        resp = self.client.get(reverse("query_create"))
        self.assertTemplateUsed(resp, 'explorer/query.html')
        self.assertContains(resp, "New Query")


def custom_view(request):
    return redirect('/custom/login')


class TestQueryDetailView(TestCase):
    databases = ['default', 'alt']

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    def test_query_with_bad_sql_renders_error(self):
        query = SimpleQueryFactory(sql="error")
        resp = self.client.get(
            reverse("query_detail", kwargs={'query_id': query.id})
        )
        self.assertTemplateUsed(resp, 'explorer/query.html')
        self.assertContains(resp, "syntax error")

    def test_query_with_bad_sql_renders_error_on_save(self):
        query = SimpleQueryFactory(sql="select 1;")
        resp = self.client.post(
            reverse("query_detail", kwargs={'query_id': query.id}),
            data={'sql': 'error'}
        )
        self.assertTemplateUsed(resp, 'explorer/query.html')
        self.assertContains(resp, "syntax error")

    def test_posting_query_saves_correctly(self):
        expected = 'select 2;'
        query = SimpleQueryFactory(sql="select 1;")
        data = model_to_dict(query)
        data['sql'] = expected
        self.client.post(
            reverse("query_detail", kwargs={'query_id': query.id}),
            data
        )
        self.assertEqual(Query.objects.get(pk=query.id).sql, expected)

    def test_change_permission_required_to_save_query(self):
        query = SimpleQueryFactory()
        expected = query.sql
        resp = self.client.get(
            reverse("query_detail", kwargs={'query_id': query.id})
        )
        self.assertTemplateUsed(resp, 'explorer/query.html')

        self.client.post(
            reverse("query_detail", kwargs={'query_id': query.id}),
            {'sql': 'select 1;'}
        )
        self.assertEqual(Query.objects.get(pk=query.id).sql, expected)

    def test_modified_date_gets_updated_after_viewing_query(self):
        query = SimpleQueryFactory()
        old = query.last_run_date
        time.sleep(0.1)
        self.client.get(
            reverse("query_detail", kwargs={'query_id': query.id})
        )
        self.assertNotEqual(old, Query.objects.get(pk=query.id).last_run_date)

    def test_doesnt_render_results_if_show_is_none(self):
        query = SimpleQueryFactory(sql='select 6870+1;')
        resp = self.client.get(
            reverse(
                "query_detail", kwargs={'query_id': query.id}
            ) + '?show=0'
        )
        self.assertTemplateUsed(resp, 'explorer/query.html')
        self.assertNotContains(resp, '6871')

    def test_doesnt_render_results_if_show_is_none_on_post(self):
        query = SimpleQueryFactory(sql='select 6870+1;')
        resp = self.client.post(
            reverse(
                "query_detail", kwargs={'query_id': query.id}
            ) + '?show=0',
            {'sql': 'select 6870+2;'}
        )
        self.assertTemplateUsed(resp, 'explorer/query.html')
        self.assertNotContains(resp, '6872')

    def test_doesnt_render_results_if_params_and_no_autorun(self):
        with self.settings(EXPLORER_AUTORUN_QUERY_WITH_PARAMS=False):
            reload_app_settings()
            query = SimpleQueryFactory(sql='select 6870+3 where 1=$$myparam:1$$;')
            resp = self.client.get(
                reverse(
                    "query_detail", kwargs={'query_id': query.id}
                )
            )
            self.assertTemplateUsed(resp, 'explorer/query.html')
            self.assertNotContains(resp, '6873')

    def test_does_render_results_if_params_and_autorun(self):
        with self.settings(EXPLORER_AUTORUN_QUERY_WITH_PARAMS=True):
            reload_app_settings()
            query = SimpleQueryFactory(sql='select 6870+4 where 1=$$myparam:1$$;')
            resp = self.client.get(
                reverse(
                    "query_detail", kwargs={'query_id': query.id}
                )
            )
            self.assertTemplateUsed(resp, 'explorer/query.html')
            self.assertContains(resp, '6874')

    def test_does_render_label_if_params_and_autorun(self):
        with self.settings(EXPLORER_AUTORUN_QUERY_WITH_PARAMS=True):
            reload_app_settings()
            query = SimpleQueryFactory(sql='select 6870+4 where 1=$$myparam|test my param label:1$$;')
            resp = self.client.get(
                reverse(
                    "query_detail", kwargs={'query_id': query.id}
                )
            )
            self.assertTemplateUsed(resp, 'explorer/query.html')
            self.assertContains(resp, 'test my param label')

    def test_admin_required(self):
        self.client.logout()
        query = SimpleQueryFactory()
        resp = self.client.get(
            reverse("query_detail", kwargs={'query_id': query.id})
        )
        self.assertTemplateUsed(resp, 'admin/login.html')

    def test_admin_required_with_explorer_no_permission_setting(self):
        self.client.logout()
        query = SimpleQueryFactory()
        with self.settings(EXPLORER_NO_PERMISSION_VIEW='explorer.tests.test_views.custom_view'):
            resp = self.client.get(
                reverse("query_detail", kwargs={'query_id': query.id})
            )
            self.assertRedirects(
                resp, '/custom/login',
                target_status_code=404
            )

    def test_individual_view_permission(self):
        self.client.logout()
        user = User.objects.create_user('user1', 'user@user.com', 'pwd')
        self.client.login(username='user1', password='pwd')

        query = SimpleQueryFactory(sql="select 123+1")

        with self.settings(EXPLORER_USER_QUERY_VIEWS={user.id: [query.id]}):
            resp = self.client.get(
                reverse("query_detail", kwargs={'query_id': query.id})
            )
        self.assertTemplateUsed(resp, 'explorer/query.html')
        self.assertContains(resp, "124")

    def test_header_token_auth(self):
        self.client.logout()

        query = SimpleQueryFactory(sql="select 123+1")

        with self.settings(EXPLORER_TOKEN_AUTH_ENABLED=True):
            resp = self.client.get(
                reverse("query_detail", kwargs={'query_id': query.id}),
                **{'HTTP_X_API_TOKEN': EXPLORER_TOKEN}
            )
        self.assertTemplateUsed(resp, 'explorer/query.html')
        self.assertContains(resp, "124")

    def test_url_token_auth(self):
        self.client.logout()

        query = SimpleQueryFactory(sql="select 123+1")

        with self.settings(EXPLORER_TOKEN_AUTH_ENABLED=True):
            resp = self.client.get(
                reverse(
                    "query_detail", kwargs={'query_id': query.id}
                ) + f'?token={EXPLORER_TOKEN}'
            )
        self.assertTemplateUsed(resp, 'explorer/query.html')
        self.assertContains(resp, "124")

    def test_user_query_views(self):
        request = Mock()

        request.user.is_anonymous = True
        kwargs = {}
        self.assertFalse(user_can_see_query(request, **kwargs))

        request.user.is_anonymous = True
        self.assertFalse(user_can_see_query(request, **kwargs))

        kwargs = {'query_id': 123}
        request.user.is_anonymous = False
        self.assertFalse(user_can_see_query(request, **kwargs))

        request.user.id = 99
        with self.settings(EXPLORER_USER_QUERY_VIEWS={99: [111, 123]}):
            self.assertTrue(user_can_see_query(request, **kwargs))

    @unittest.skipIf(not app_settings.ENABLE_TASKS, 'tasks not enabled')
    @patch('explorer.models.get_s3_bucket')
    def test_query_snapshot_renders(self, mocked_conn):
        conn = Mock()
        conn.objects.filter = Mock()
        k1 = Mock()
        k1.generate_url.return_value = 'http://s3.com/foo'
        k1.last_modified = '2015-01-01'
        k2 = Mock()
        k2.generate_url.return_value = 'http://s3.com/bar'
        k2.last_modified = '2015-01-02'
        conn.objects.filter.return_value = [k1, k2]
        mocked_conn.return_value = conn

        query = SimpleQueryFactory(sql="select 1;", snapshot=True)
        resp = self.client.get(
            reverse("query_detail", kwargs={'query_id': query.id})
        )
        self.assertContains(resp, '2015-01-01')
        self.assertContains(resp, '2015-01-02')

    def test_failing_blacklist_means_query_doesnt_execute(self):
        conn = connections[CONN]
        start = len(conn.queries)
        query = SimpleQueryFactory(sql="select 1;")
        resp = self.client.post(
            reverse("query_detail", kwargs={'query_id': query.id}),
            data={'sql': "delete from auth_user;"}
        )
        end = len(conn.queries)

        self.assertTemplateUsed(resp, 'explorer/query.html')
        self.assertContains(resp, MSG_FAILED_BLACKLIST % '')

        self.assertEqual(start, end)

    def test_fullscreen(self):
        query = SimpleQueryFactory(sql="select 1;")
        resp = self.client.get(
            reverse(
                "query_detail", kwargs={'query_id': query.id}
            ) + '?fullscreen=1'
        )
        self.assertTemplateUsed(resp, 'explorer/fullscreen.html')

    def test_multiple_connections_integration(self):
        from explorer.app_settings import EXPLORER_CONNECTIONS
        from explorer.connections import connections

        c1_alias = EXPLORER_CONNECTIONS['SQLite']
        conn = connections[c1_alias]
        c = conn.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS animals (name text NOT NULL);')
        c.execute('INSERT INTO animals ( name ) VALUES (\'peacock\')')

        c2_alias = EXPLORER_CONNECTIONS['Another']
        conn = connections[c2_alias]
        c = conn.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS animals (name text NOT NULL);')
        c.execute('INSERT INTO animals ( name ) VALUES (\'superchicken\')')

        query = SimpleQueryFactory(
            sql="select name from animals;", connection=c1_alias
        )
        resp = self.client.get(
            reverse("query_detail", kwargs={'query_id': query.id})
        )
        self.assertContains(resp, "peacock")

        query = SimpleQueryFactory(
            sql="select name from animals;", connection=c2_alias
        )
        resp = self.client.get(
            reverse("query_detail", kwargs={'query_id': query.id})
        )
        self.assertContains(resp, "superchicken")


class TestDownloadView(TestCase):
    def setUp(self):
        self.query = SimpleQueryFactory(sql="select 1;")
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    def test_admin_required(self):
        self.client.logout()
        resp = self.client.get(
            reverse("download_query", kwargs={'query_id': self.query.id})
        )
        self.assertTemplateUsed(resp, 'admin/login.html')

    def test_params_in_download(self):
        q = SimpleQueryFactory(sql="select '$$foo$$';")
        url = '{}?params={}'.format(
            reverse("download_query", kwargs={'query_id': q.id}),
            'foo:123'
        )
        resp = self.client.get(url)
        self.assertContains(resp, "'123'")

    def test_download_defaults_to_csv(self):
        query = SimpleQueryFactory()
        url = reverse("download_query", args=[query.pk])

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['content-type'], 'text/csv')

    def test_download_csv(self):
        query = SimpleQueryFactory()
        url = reverse("download_query", args=[query.pk]) + '?format=csv'

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['content-type'], 'text/csv')

    def test_bad_query_gives_500(self):
        query = SimpleQueryFactory(sql='bad')
        url = reverse("download_query", args=[query.pk]) + '?format=csv'

        response = self.client.get(url)

        self.assertEqual(response.status_code, 500)

    def test_download_json(self):
        query = SimpleQueryFactory()
        url = reverse("download_query", args=[query.pk]) + '?format=json'

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['content-type'], 'application/json')

        json_data = json.loads(response.content.decode('utf-8'))
        self.assertIsInstance(json_data, list)
        self.assertEqual(len(json_data), 1)
        self.assertEqual(json_data, [{'TWO': 2}])


class TestQueryPlayground(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    def test_empty_playground_renders(self):
        resp = self.client.get(reverse("explorer_playground"))
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed(resp, 'explorer/play.html')

    def test_playground_renders_with_query_sql(self):
        query = SimpleQueryFactory(sql="select 1;")
        resp = self.client.get(
            '{}?query_id={}'.format(reverse("explorer_playground"), query.id)
        )
        self.assertTemplateUsed(resp, 'explorer/play.html')
        self.assertContains(resp, 'select 1;')

    def test_playground_renders_with_posted_sql(self):
        resp = self.client.post(
            reverse("explorer_playground"),
            {'sql': 'select 1+3400;'}
        )
        self.assertTemplateUsed(resp, 'explorer/play.html')
        self.assertContains(resp, '3401')

    def test_playground_doesnt_render_with_posted_sql_if_show_is_none(self):
        resp = self.client.post(
            reverse("explorer_playground") + '?show=0',
            {'sql': 'select 1+3400;'}
        )
        self.assertTemplateUsed(resp, 'explorer/play.html')
        self.assertNotContains(resp, '3401')

    def test_playground_renders_with_empty_posted_sql(self):
        resp = self.client.post(reverse("explorer_playground"), {'sql': ''})
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed(resp, 'explorer/play.html')

    def test_query_with_no_resultset_doesnt_throw_error(self):
        query = SimpleQueryFactory(sql="")
        resp = self.client.get(
            '{}?query_id={}'.format(reverse("explorer_playground"), query.id)
        )
        self.assertTemplateUsed(resp, 'explorer/play.html')

    def test_admin_required(self):
        self.client.logout()
        resp = self.client.get(reverse("explorer_playground"))
        self.assertTemplateUsed(resp, 'admin/login.html')

    def test_admin_required_with_no_permission_view_setting(self):
        self.client.logout()
        with self.settings(EXPLORER_NO_PERMISSION_VIEW='explorer.tests.test_views.custom_view'):
            resp = self.client.get(reverse("explorer_playground"))
            self.assertRedirects(
                resp,
                '/custom/login',
                target_status_code=404
            )

    def test_loads_query_from_log(self):
        querylog = QueryLogFactory()
        resp = self.client.get(
            '{}?querylog_id={}'.format(
                reverse("explorer_playground"), querylog.id
            )
        )
        self.assertContains(resp, "FOUR")

    def test_fails_blacklist(self):
        resp = self.client.post(
            reverse("explorer_playground"),
            {'sql': "delete from auth_user;"}
        )
        self.assertTemplateUsed(resp, 'explorer/play.html')
        self.assertContains(resp, MSG_FAILED_BLACKLIST % '')

    def test_fullscreen(self):
        query = SimpleQueryFactory(sql="")
        resp = self.client.get(
            '{}?query_id={}&fullscreen=1'.format(
                reverse("explorer_playground"),
                query.id
            )
        )
        self.assertTemplateUsed(resp, 'explorer/fullscreen.html')


class TestCSVFromSQL(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    def test_admin_required(self):
        self.client.logout()
        resp = self.client.post(reverse("download_sql"), {})
        self.assertTemplateUsed(resp, 'admin/login.html')

    def test_downloading_from_playground(self):
        sql = "select 1;"
        resp = self.client.post(reverse("download_sql"), {'sql': sql})
        self.assertIn('attachment', resp['Content-Disposition'])
        self.assertEqual('text/csv', resp['content-type'])
        ql = QueryLog.objects.first()
        self.assertIn(
            'filename="Playground-%s.csv"' % ql.id,
            resp['Content-Disposition']
        )

    def test_stream_csv_from_query(self):
        q = SimpleQueryFactory()
        resp = self.client.get(
            reverse("stream_query", kwargs={'query_id': q.id})
        )
        self.assertEqual('text/csv', resp['content-type'])


class TestSQLDownloadViews(TestCase):
    databases = ['default', 'alt']

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    def test_sql_download_csv(self):
        url = reverse("download_sql") + '?format=csv'

        response = self.client.post(url, {'sql': 'select 1;'})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['content-type'], 'text/csv')

    def test_sql_download_respects_connection(self):
        from explorer.app_settings import EXPLORER_CONNECTIONS
        from explorer.connections import connections

        c1_alias = EXPLORER_CONNECTIONS['SQLite']
        conn = connections[c1_alias]
        c = conn.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS animals (name text NOT NULL);')
        c.execute('INSERT INTO animals ( name ) VALUES (\'peacock\')')

        c2_alias = EXPLORER_CONNECTIONS['Another']
        conn = connections[c2_alias]
        c = conn.cursor()
        c.execute('CREATE TABLE IF NOT EXISTS animals (name text NOT NULL);')
        c.execute('INSERT INTO animals ( name ) VALUES (\'superchicken\')')

        url = reverse("download_sql") + '?format=csv'

        response = self.client.post(
            url,
            {'sql': 'select * from animals;', 'connection': c2_alias}
        )

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'superchicken')

    def test_sql_download_csv_with_custom_delim(self):
        url = reverse("download_sql") + '?format=csv&delim=|'

        response = self.client.post(url, {'sql': 'select 1,2;'})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['content-type'], 'text/csv')
        self.assertEqual(response.content.decode('utf-8-sig'), '1|2\r\n1|2\r\n')

    def test_sql_download_csv_with_tab_delim(self):
        url = reverse("download_sql") + '?format=csv&delim=tab'

        response = self.client.post(url, {'sql': 'select 1,2;'})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['content-type'], 'text/csv')
        self.assertEqual(response.content.decode('utf-8-sig'), '1\t2\r\n1\t2\r\n')

    def test_sql_download_csv_with_bad_delim(self):
        url = reverse("download_sql") + '?format=csv&delim=foo'

        response = self.client.post(url, {'sql': 'select 1,2;'})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['content-type'], 'text/csv')
        self.assertEqual(response.content.decode('utf-8-sig'), '1,2\r\n1,2\r\n')

    def test_sql_download_json(self):
        url = reverse("download_sql") + '?format=json'

        response = self.client.post(url, {'sql': 'select 1;'})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['content-type'], 'application/json')


class TestSchemaView(TestCase):

    def setUp(self):
        cache.clear()
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    def test_returns_schema_contents(self):
        resp = self.client.get(
            reverse("explorer_schema", kwargs={'connection': CONN})
        )
        self.assertContains(resp, "explorer_query")
        self.assertTemplateUsed(resp, 'explorer/schema.html')

    def test_returns_404_if_conn_doesnt_exist(self):
        resp = self.client.get(
            reverse("explorer_schema", kwargs={'connection': 'foo'})
        )
        self.assertEqual(resp.status_code, 404)

    def test_admin_required(self):
        self.client.logout()
        resp = self.client.get(
            reverse("explorer_schema", kwargs={'connection': CONN})
        )
        self.assertTemplateUsed(resp, 'admin/login.html')

    @unittest.skipIf(not app_settings.ENABLE_TASKS, 'tasks not enabled')
    @patch('explorer.schema.do_async')
    def test_builds_async(self, mocked_async_check):
        mocked_async_check.return_value = True
        resp = self.client.get(
            reverse("explorer_schema", kwargs={'connection': CONN})
        )
        self.assertTemplateUsed(resp, 'explorer/schema_building.html')
        resp = self.client.get(
            reverse("explorer_schema", kwargs={'connection': CONN})
        )
        self.assertTemplateUsed(resp, 'explorer/schema.html')


class TestFormat(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    def test_returns_formatted_sql(self):
        resp = self.client.post(
            reverse("format_sql"),
            data={"sql": "select * from explorer_query"}
        )
        resp = json.loads(resp.content.decode('utf-8'))
        self.assertIn("\n", resp['formatted'])
        self.assertIn("explorer_query", resp['formatted'])


class TestParamsInViews(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')
        self.query = SimpleQueryFactory(sql="select $$swap$$;")

    def test_retrieving_query_works_with_params(self):
        resp = self.client.get(
            reverse(
                "query_detail", kwargs={'query_id': self.query.id}
            ) + '?params=swap:123}'
        )
        self.assertContains(resp, "123")

    def test_saving_non_executing_query_with__wrong_url_params_works(self):
        q = SimpleQueryFactory(sql="select $$swap$$;")
        data = model_to_dict(q)
        url = '{}?params={}'.format(
            reverse("query_detail", kwargs={'query_id': q.id}),
            'foo:123'
        )
        resp = self.client.post(url, data)
        self.assertContains(resp, 'saved')

    def test_users_without_change_permissions_can_use_params(self):
        resp = self.client.get(
            reverse(
                "query_detail", kwargs={'query_id': self.query.id}
            ) + '?params=swap:123}'
        )
        self.assertContains(resp, "123")


class TestCreatedBy(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.user2 = User.objects.create_superuser(
            'admin2', 'admin2@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')
        self.query = SimpleQueryFactory.build(created_by_user=self.user)
        self.data = model_to_dict(self.query)
        del self.data['id']
        self.data["created_by_user_id"] = self.user2.id

    def test_query_update_doesnt_change_created_user(self):
        self.query.save()
        self.client.post(
            reverse("query_detail", kwargs={'query_id': self.query.id}),
            self.data
        )
        q = Query.objects.get(id=self.query.id)
        self.assertEqual(q.created_by_user_id, self.user.id)

    def test_new_query_gets_created_by_logged_in_user(self):
        self.client.post(reverse("query_create"), self.data)
        q = Query.objects.first()
        self.assertEqual(q.created_by_user_id, self.user.id)


class TestQueryLog(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    def test_playground_saves_query_to_log(self):
        self.client.post(reverse("explorer_playground"), {'sql': 'select 1;'})
        log = QueryLog.objects.first()
        self.assertTrue(log.is_playground)
        self.assertEqual(log.sql, 'select 1;')

    # Since it will be saved on the initial query creation, no need to log it
    def test_creating_query_does_not_save_to_log(self):
        query = SimpleQueryFactory()
        self.client.post(reverse("query_create"), model_to_dict(query))
        self.assertEqual(0, QueryLog.objects.count())

    def test_query_saves_to_log(self):
        query = SimpleQueryFactory()
        data = model_to_dict(query)
        data['sql'] = 'select 12345;'
        self.client.post(
            reverse("query_detail", kwargs={'query_id': query.id}),
            data
        )
        self.assertEqual(1, QueryLog.objects.count())

    def test_query_gets_logged_and_appears_on_log_page(self):
        query = SimpleQueryFactory()
        data = model_to_dict(query)
        data['sql'] = 'select 12345;'
        self.client.post(
            reverse("query_detail", kwargs={'query_id': query.id}),
            data
        )
        resp = self.client.get(reverse("explorer_logs"))
        self.assertContains(resp, 'select 12345;')

    def test_admin_required(self):
        self.client.logout()
        resp = self.client.get(reverse("explorer_logs"))
        self.assertTemplateUsed(resp, 'admin/login.html')

    def test_is_playground(self):
        self.assertTrue(QueryLog(sql='foo').is_playground)

        q = SimpleQueryFactory()
        self.assertFalse(QueryLog(sql='foo', query_id=q.id).is_playground)


class TestEmailQuery(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')

    @patch('explorer.views.email.execute_query')
    def test_email_calls_task(self, mocked_execute):
        query = SimpleQueryFactory()
        url = reverse("email_csv_query", kwargs={'query_id': query.id})
        self.client.post(
            url,
            data={'email': 'foo@bar.com'},
            **{'HTTP_X_REQUESTED_WITH': 'XMLHttpRequest'}
        )
        self.assertEqual(mocked_execute.delay.call_count, 1)

    def test_email_403(self):
        query = SimpleQueryFactory()
        url = reverse("email_csv_query", kwargs={'query_id': query.id})
        response = self.client.post(
            url,
            data={},
            **{'HTTP_X_REQUESTED_WITH': 'XMLHttpRequest'}
        )
        self.assertEqual(response.status_code, 403)

    def test_email_no_xml_403(self):
        query = SimpleQueryFactory()
        url = reverse("email_csv_query", kwargs={'query_id': query.id})
        response = self.client.post(
            url,
            data={},
        )
        self.assertEqual(response.status_code, 403)


class TestQueryFavorites(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')
        self.q = SimpleQueryFactory(title='query for x, y')
        QueryFavorite.objects.create(user=self.user, query=self.q)

    def test_returns_favorite_list(self):
        resp = self.client.get(
            reverse("query_favorites")
        )
        self.assertContains(resp, 'query for x, y')


class TestQueryFavorite(TestCase):

    def setUp(self):
        self.user = User.objects.create_superuser(
            'admin', 'admin@admin.com', 'pwd'
        )
        self.client.login(username='admin', password='pwd')
        self.q = SimpleQueryFactory(title='query for x, y')

    def test_toggle(self):
        resp = self.client.post(
            reverse("query_favorite",  args=(self.q.id,))
        )
        resp = json.loads(resp.content.decode('utf-8'))
        self.assertTrue(resp['is_favorite'])
        resp = self.client.post(
            reverse("query_favorite",  args=(self.q.id,))
        )
        resp = json.loads(resp.content.decode('utf-8'))
        self.assertFalse(resp['is_favorite'])



from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.views import LoginView
from django.core.exceptions import ImproperlyConfigured

from explorer import app_settings, permissions


class PermissionRequiredMixin:

    permission_required = None

    @staticmethod
    def handle_no_permission(request):
        return app_settings.EXPLORER_NO_PERMISSION_VIEW()(request)

    def get_permission_required(self):
        if self.permission_required is None:
            raise ImproperlyConfigured(
                '{0} is missing the permission_required attribute. '
                'Define {0}.permission_required, or override '
                '{0}.get_permission_required().'.format(
                    self.__class__.__name__
                )
            )
        return self.permission_required

    def has_permission(self, request, *args, **kwargs):
        perms = self.get_permission_required()

        # TODO: fix the case when the perms is not defined in
        #  permissions module.
        handler = getattr(permissions, perms)
        return handler(request, *args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        if not self.has_permission(request, *args, **kwargs):
            return self.handle_no_permission(request)
        return super().dispatch(request, *args, **kwargs)


class SafeLoginView(LoginView):
    template_name = 'admin/login.html'


def safe_login_view_wrapper(request):
    return SafeLoginView.as_view(
        extra_context={
            'title': 'Log in',
            REDIRECT_FIELD_NAME: request.get_full_path()
        }
    )(request)



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('explorer', '0005_auto_20160105_2052'),
    ]

    operations = [
        migrations.AddField(
            model_name='query',
            name='connection',
            field=models.CharField(help_text=b'Name of DB connection (as specified in settings) to use for this query. Will use EXPLORER_DEFAULT_CONNECTION if left blank', max_length=128, null=True, blank=True),
        ),
    ]



from django.core.cache import cache

from explorer.app_settings import (
    ENABLE_TASKS, EXPLORER_ASYNC_SCHEMA, EXPLORER_CONNECTIONS, EXPLORER_SCHEMA_EXCLUDE_TABLE_PREFIXES,
    EXPLORER_SCHEMA_INCLUDE_TABLE_PREFIXES, EXPLORER_SCHEMA_INCLUDE_VIEWS,
)
from explorer.tasks import build_schema_cache_async
from explorer.utils import get_valid_connection


# These wrappers make it easy to mock and test
def _get_includes():
    return EXPLORER_SCHEMA_INCLUDE_TABLE_PREFIXES


def _get_excludes():
    return EXPLORER_SCHEMA_EXCLUDE_TABLE_PREFIXES


def _include_views():
    return EXPLORER_SCHEMA_INCLUDE_VIEWS is True


def do_async():
    return ENABLE_TASKS and EXPLORER_ASYNC_SCHEMA


def _include_table(t):
    if _get_includes() is not None:
        return any([t.startswith(p) for p in _get_includes()])
    return not any([t.startswith(p) for p in _get_excludes()])


def connection_schema_cache_key(connection_alias):
    return f'_explorer_cache_key_{connection_alias}'


def schema_info(connection_alias):
    key = connection_schema_cache_key(connection_alias)
    ret = cache.get(key)
    if ret:
        return ret
    if do_async():
        build_schema_cache_async.delay(connection_alias)
    else:
        return build_schema_cache_async(connection_alias)


def build_schema_info(connection_alias):
    """
        Construct schema information via engine-specific queries of the
        tables in the DB.

        :return: Schema information of the following form,
                 sorted by db_table_name.
            [
                ("db_table_name",
                    [
                        ("db_column_name", "DbFieldType"),
                        (...),
                    ]
                )
            ]

        """
    connection = get_valid_connection(connection_alias)
    ret = []
    with connection.cursor() as cursor:
        tables_to_introspect = connection.introspection.table_names(
            cursor, include_views=_include_views()
        )

        for table_name in tables_to_introspect:
            if not _include_table(table_name):
                continue
            td = []
            table_description = connection.introspection.get_table_description(
                cursor, table_name
            )
            for row in table_description:
                column_name = row[0]
                try:
                    field_type = connection.introspection.get_field_type(
                        row[1], row
                    )
                except KeyError:
                    field_type = 'Unknown'
                td.append((column_name, field_type))
            ret.append((table_name, td))
    return ret


def build_async_schemas():
    if do_async():
        for c in EXPLORER_CONNECTIONS:
            schema_info(c)



import os
import sys
from pathlib import Path

from setuptools import setup


try:
    from sphinx.setup_command import BuildDoc
except ImportError:
    BuildDoc = None

from explorer import get_version


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...

name = "django-sql-explorer"
version = get_version(True)
release = get_version()


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


if sys.argv[-1] == 'build':
    os.system('python setup.py sdist bdist_wheel')
    print(f"Built release {release} (version {version})")
    sys.exit()

if sys.argv[-1] == 'release':
    os.system('twine upload --skip-existing dist/*')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version:")
    os.system(f"git tag -a {version} -m 'version {version}'")
    os.system("git push --tags")
    sys.exit()

this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()

setup(
    name=name,
    version=version,
    author="Chris Clark",
    author_email="chris@untrod.com",
    maintainer='Mark Walker',
    maintainer_email='theshow@gmail.com',
    description=("A pluggable app that allows users (admins) to execute SQL,"
                 " view, and export the results."),
    license="MIT",
    keywords="django sql explorer reports reporting csv database query",
    url="https://github.com/groveco/django-sql-explorer",
    project_urls={
      'Changes': 'https://django-sql-explorer.readthedocs.io/en/latest/history.html',
      'Documentation': 'https://django-sql-explorer.readthedocs.io/en/latest/',
      'Issues': 'https://github.com/groveco/django-sql-explorer/issues'
    },
    packages=['explorer'],
    long_description=long_description,
    long_description_content_type='text/x-rst',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Utilities',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.1',
        'Framework :: Django :: 4.2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3 :: Only',
    ],
    python_requires='>=3.8',
    install_requires=[
        'Django>=3.2.12',
        'sqlparse>=0.4.0',
    ],
    extras_require={
        "charts": [
            "matplotlib<4",
            "seaborn<0.12"
        ],
        "snapshots": [
            "boto3>=1.20.0",
            "celery>=4.0"
        ],
        "xls": [
            'xlsxwriter>=1.3.6'
        ]
    },
    cmdclass={
        'build_sphinx': BuildDoc,
    },
    command_options={
        'build_sphinx': {
            'project': ('setup.py', name),
            'version': ('setup.py', version),
            'release': ('setup.py', release),
            'source_dir': ('setup.py', 'docs'),
            'build_dir': ('setup.py', './docs/_build')
        }
    },
    include_package_data=True,
    zip_safe=False,
)



from django.urls import path

from explorer.views import (
    CreateQueryView, DeleteQueryView, DownloadFromSqlView, DownloadQueryView, EmailCsvQueryView, ListQueryLogView,
    ListQueryView, PlayQueryView, QueryFavoritesView, QueryFavoriteView, QueryView, SchemaView, StreamQueryView,
    format_sql,
)


urlpatterns = [
    path(
        '<int:query_id>/', QueryView.as_view(), name='query_detail'
    ),
    path(
        '<int:query_id>/download', DownloadQueryView.as_view(),
        name='download_query'
    ),
    path(
        '<int:query_id>/stream', StreamQueryView.as_view(),
        name='stream_query'
    ),
    path('download', DownloadFromSqlView.as_view(), name='download_sql'),
    path(
        '<int:query_id>/email_csv', EmailCsvQueryView.as_view(),
        name='email_csv_query'
    ),
    path(
        '<int:pk>/delete', DeleteQueryView.as_view(), name='query_delete'
    ),
    path('new/', CreateQueryView.as_view(), name='query_create'),
    path('play/', PlayQueryView.as_view(), name='explorer_playground'),
    path(
        'schema/<path:connection>', SchemaView.as_view(),
        name='explorer_schema'
    ),
    path('logs/', ListQueryLogView.as_view(), name='explorer_logs'),
    path('format/', format_sql, name='format_sql'),
    path('favorites/', QueryFavoritesView.as_view(), name='query_favorites'),
    path('favorite/<int:query_id>', QueryFavoriteView.as_view(), name='query_favorite'),
    path('', ListQueryView.as_view(), name='explorer_index'),
]



from django.test import TestCase, override_settings


try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from django.conf import settings
from django.contrib.auth.models import User


class TestCsrfCookieName(TestCase):
    def test_csrf_cookie_name_in_context(self):
        self.user = User.objects.create_superuser('admin', 'admin@admin-fake.com', 'pwd')
        self.client.login(username='admin', password='pwd')
        resp = self.client.get(reverse('explorer_index'))
        self.assertTrue('csrf_cookie_name' in resp.context)
        self.assertEqual(resp.context['csrf_cookie_name'], settings.CSRF_COOKIE_NAME)

    @override_settings(CSRF_COOKIE_NAME='TEST_CSRF_COOKIE_NAME')
    def test_custom_csrf_cookie_name(self):
        self.user = User.objects.create_superuser('admin', 'admin@admin-fake.com', 'pwd')
        self.client.login(username='admin', password='pwd')
        resp = self.client.get(reverse('explorer_index'))
        self.assertTrue('csrf_cookie_name' in resp.context)
        self.assertEqual(resp.context['csrf_cookie_name'], 'TEST_CSRF_COOKIE_NAME')
